#include <iostream>
#include "shapefil.h"
#include <time.h>
#include <fstream>
#include <string>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

//#define _CRT_SECURE_NO_WARNINGS

using namespace std;

void printUsage(void) {
	printf("[INF] Usage: GPSFDshp2mp <selected shape layer> <type column> <label column> (<include all columns (default yes)>) (<dictionary file>) (<mapid>) (<MapName>)\n");
}


int main( int argc, char *argv[])
{
	clock_t start, end;
	start = clock();

	DBFFieldType  dbfType;

	int nWidth, nDecimals;
	SHPObject *obj;


	//info only
	printf("----------------------------------------------------------------------\n");
	printf("| Welcome to GPSFileDepot shp2mp                                     |\n");
	printf("| This was modified from sprok's shp2mp                              |\n");
	printf("| Version: 0.9.8  (20160229)                                         |\n");
	printf("| Licenses: GPLv3, LGPLv2                                            |\n");
	printf("| Updates/info: http://www.gpsfiledepot.com/tools/GPSFDshp2mp.php    |\n");
	printf("|                                                                    |\n");
	printf("|                          ---USAGE---                               |\n");
	printf("|                                                                    |\n");
	printf("| Inputs: shapefile.shp type label (include all y/n) (dictionary.mp) |\n");
	printf("| (MapID) (MapName)                                                  |\n");
	printf("| Type must be the column with type information and label the column |\n");
	printf("| with name information. Include all should be n for only output the |\n");
	printf("| geographic data, name, and type code. Map ID & name can only be    |\n");
	printf("| used with a custom dictionary.                                     |\n");
	printf("----------------------------------------------------------------------\n\n");

	if (argc != 4 && argc != 5 && argc != 6 && argc != 7 && argc != 8) {
		printf("[ERR] Incorrect parameter! Exiting...\n");
		printUsage();
		return 0;
	}

	DBFHandle currentDBFFile = DBFOpen(argv[1], "rb" ); //Opening .dbf and .shp file
	SHPHandle currentSHPFile = SHPOpen(argv[1], "rb" );

	if (!currentDBFFile) {
		printf("[ERR] No .dbf file found. Exiting...\n");
		return 0;
	}
	if (!currentSHPFile) {
		printf("[ERR] No .shp file found. Exiting...\n");
		return 0;
	}

	
	FILE *outputFile;
	string outputFileName = "";
	outputFileName.append(argv[1]);
	outputFileName.append(".mp");
	outputFile = fopen( outputFileName.c_str(),"w");

	if (!argv[5]) {
		printf("[INF] No dictionary file provided, using default\n");
		fprintf(outputFile,
			";#################\n"
			";Created by GPSFileDepot shp2mp\n"
			";#################\n"
			";\n"
			"[IMG ID]\n"
			"ID=99999999\n"
			"Name=test_map\n"
			"Elevation=M\n"
			"LBLcoding=9\n"
			"Codepage=1250\n"
			"Marine=N\n"
			"Copyright=Check the Copyright details for the source file!\n"
			"Preprocess=F\n"
			";TreSize=3000\n"
			"POIIndex=N\n"
			"Transparent=Y\n"
			"Levels=4\n"
			"Level0=24\n"
			"Level1=22\n"
			"Level2=20\n"
			"Level3=18\n"
			"Zoom0=1\n"
			"Zoom1=2\n"
			"Zoom2=3\n"
			"Zoom3=4\n"
			"[END-IMG ID]\n\n");
	} //no dictionary file
	else {
		cout << "[INF] Using dictionary file: " << argv[5] << '\n';
		string dictionaryLine;
		ifstream dictionaryFile(argv[5]);
		if (dictionaryFile.is_open()) {
			while (getline(dictionaryFile, dictionaryLine)) {
				if (argc > 6 && strcmp(dictionaryLine.c_str(), "ID=") == 0) {
					fprintf(outputFile, "ID=");
					fprintf(outputFile, argv[6]);
					cout << "[INF] Using custom ID: " << argv[6] << '\n';
				}
				else if (argc > 7 && strcmp(dictionaryLine.c_str(), "Name=") == 0) {
					fprintf(outputFile, "Name=");
					fprintf(outputFile, argv[7]);
					cout << "[INF] Using custom map name: " << argv[7] << '\n';
				}
				else {
					fprintf(outputFile, dictionaryLine.c_str());
				}
				fprintf(outputFile, "\n");
			}
			dictionaryFile.close();
		}
		else {
			printf("[ERR] Dictionary file not found. Exiting...\n");
			return 0;
		}
	} //dictionary file

	const int DBFFieldCount = DBFGetFieldCount(currentDBFFile);

	string ColumnTitle[255];

	for (int i=0; i<DBFFieldCount; ++i)	{
		char Title[50];
		dbfType = DBFGetFieldInfo(currentDBFFile, i, Title, &nWidth, &nDecimals);
		ColumnTitle[i]= Title ;
	}

	obj = SHPReadObject(currentSHPFile, 1);

	if (obj->nSHPType==0 ) {
		printf("[CRIT] The ""%s"" file contains NULL elements. Exiting...\n",argv[1]);
		return 0;
	}
	else if (obj->nSHPType==1 ) {
		printf("[INF] The ""%s"" file contains Point elements.\n",argv[1]);
	}
	else if (obj->nSHPType==3 ) {
		printf("[INF] The ""%s"" file contains Polyline elements.\n",argv[1]);
	}
	else if (obj->nSHPType==5 ) {
		printf("[INF] The ""%s"" file contains Polygon elements.\n",argv[1]);
	}
	else {
		printf("[CRIT] The ""%s"" file contains elements that currently not supported. Exiting...\n",outputFileName.c_str());
		return 0;
	}
	

	int DBFTotalElements = DBFGetRecordCount(currentDBFFile);
	if (DBFTotalElements == 0) {
		printf("[WRN] No elements in .dbf file!\n");
	}

	/////////////////////////////// Point process
	if (obj->nSHPType==1) {
		for (int i=0; i<DBFTotalElements; ++i) {
			obj = SHPReadObject(currentSHPFile, i);

			fprintf(outputFile, "[POI]\n"
				"Data0=");

			fprintf(outputFile, "(%.6f,%.6f)\n",obj->padfY[0],obj->padfX[0]);

			for (int j=0; j<DBFFieldCount; ++j) { //Print out all the columns' content
				if (ColumnTitle[j] == argv[2]) {
					fprintf(outputFile, "%s=%s\n", "Type", DBFReadStringAttribute(currentDBFFile, i, j));
				}
				else if (ColumnTitle[j] == argv[3]) {
					fprintf(outputFile, "%s=%s\n", "Label", DBFReadStringAttribute(currentDBFFile, i, j));
				}
				else {
					if (strcmp(argv[4], "n") != 0) {
						fprintf(outputFile, "%s=%s\n", ColumnTitle[j].c_str(), DBFReadStringAttribute(currentDBFFile, i, j));
					}
				}
			}

			fprintf(outputFile,"[END]\n\n");
			SHPDestroyObject(obj);

			printf("\r[INF] Processing: %.0f%%", (double)i/(double)DBFTotalElements*100 );
		}
	}


	/////////////////////////////// Polyline process
	else if (obj->nSHPType==3) {

		for (int i=0; i<DBFTotalElements; ++i) {
			obj = SHPReadObject(currentSHPFile, i);

			fprintf(outputFile, "[POLYLINE]\n"
				"Data0=");
			for (int j=0; j<obj->nVertices; ++j) {
				fprintf(outputFile, "(%.6f,%.6f)",obj->padfY[j],obj->padfX[j]);

				if ((j+1) != obj->nVertices ) {
					fprintf(outputFile,",");
				}
				else {
					fprintf(outputFile,"\n");
				}
			}

			for (int j=0; j<DBFFieldCount; ++j) {
				if (ColumnTitle[j] == argv[2]) {
					fprintf(outputFile, "%s=%s\n", "Type", DBFReadStringAttribute(currentDBFFile, i, j));
				}
				else if (ColumnTitle[j] == argv[3]) {
					fprintf(outputFile, "%s=%s\n", "Label", DBFReadStringAttribute(currentDBFFile, i, j));
				}
				else {
					if (strcmp(argv[4],"n") != 0) {
						fprintf(outputFile, "%s=%s\n", ColumnTitle[j].c_str(), DBFReadStringAttribute(currentDBFFile, i, j));
					}
				}
			}

			fprintf(outputFile,"[END]\n\n");
			SHPDestroyObject(obj);

			printf("\r[INF] Processing: %.0f%%", (double)i/(double)DBFTotalElements*100 );
		}
	}

	/////////////////////////////// Polygon process
	else if (obj->nSHPType==5) {

		for (int i=0; i<DBFTotalElements; ++i) {
			obj = SHPReadObject(currentSHPFile, i);

			fprintf(outputFile, "[POLYGON]\n"
				"Data0=");

						if (obj->nParts==1) {

							for (int j=0; j<obj->nVertices; ++j) {
								fprintf(outputFile, "(%.6f,%.6f)",obj->padfY[j],obj->padfX[j]);

								if ((j+1) != obj->nVertices ) {
									fprintf(outputFile,",");
								}
								else {
									fprintf(outputFile,"\n");
								}
							}
							
							for (int j=0; j<DBFFieldCount; ++j)	{
								if (ColumnTitle[j] == argv[2]) {
									fprintf(outputFile, "%s=%s\n", "Type", DBFReadStringAttribute(currentDBFFile, i, j));
								}
								else if (ColumnTitle[j] == argv[3]) {
									fprintf(outputFile, "%s=%s\n", "Label", DBFReadStringAttribute(currentDBFFile, i, j));
								}
								else {
									if (strcmp(argv[4], "n") != 0) {
										fprintf(outputFile, "%s=%s\n", ColumnTitle[j].c_str(), DBFReadStringAttribute(currentDBFFile, i, j));
									}
								}
							}

						//fprintf(outputFile,"mpID=%i\n",i+1);

						}
						///////multipolygon
						else if (obj->nParts > 1) {
								
								int totalNode=obj->nVertices;
								int k=0;
								int parts=obj->nParts;
								bool last = false;

								for (int pp=1; pp<=obj->nParts;++pp) {
									int Start=obj->panPartStart[pp-1];
									int End=obj->panPartStart[pp];

									if (pp==obj->nParts) {
										End = totalNode;
										last = true;
									}
									int alma=1;

									if (Start == 0)	{
										for(;;++k) {
											fprintf(outputFile, "(%.6f,%.6f)",obj->padfY[k],obj->padfX[k]);
											
											if (k!=(End-1)) {
												fprintf(outputFile,",");
											}						
											if (k==(End-1)) {
												fprintf(outputFile,"\nData0=");
												k++;
												break;
											}
										}
									}

									if (k==Start && !last) {
										for(;;++k) {
											fprintf(outputFile, "(%.6f,%.6f)",obj->padfY[k],obj->padfX[k]);
											
											if (k!=(End-1))	{
												fprintf(outputFile,",");
											}
											if (k==(End-1)) {
												fprintf(outputFile,"\nData0=");
												k++;
												break;
											}
										}
									}

									if (k==Start && last) {
										int alma=1;
										for(;k<totalNode;++k) {
											fprintf(outputFile, "(%.6f,%.6f)",obj->padfY[k],obj->padfX[k]);
																					
											if (k!=(End-1)) {
												fprintf(outputFile,",");
											}
											if (k==(End-1)) {
												fprintf(outputFile,"\n");												
												break;
											}											
										}
									}											
								}															
							
							for (int j=0; j<DBFFieldCount; ++j) {
								if (ColumnTitle[j] == argv[2]) {
									fprintf(outputFile, "%s=%s\n", "Type", DBFReadStringAttribute(currentDBFFile, i, j));
								}
								else if (ColumnTitle[j] == argv[3]) {
									fprintf(outputFile, "%s=%s\n", "Label", DBFReadStringAttribute(currentDBFFile, i, j));
								}
								else {
									if (strcmp(argv[4], "n") != 0) {
										fprintf(outputFile, "%s=%s\n", ColumnTitle[j].c_str(), DBFReadStringAttribute(currentDBFFile, i, j));
									}
								}
							}
							fprintf(outputFile,"mpID=%i\n",i+1);
						}					

			fprintf(outputFile,"[END]\n\n");
			SHPDestroyObject(obj);

			printf("\r[INF] Processing: %.0f%%", (double)i/(double)DBFTotalElements*100 );
		}
	}
	else {
		printf("[CRIT] This shp file contains elements that are currently not supported. Exiting...\n");
		return 0;
	}

	fclose (outputFile);
	DBFClose (currentDBFFile);
	SHPClose (currentSHPFile);

	printf("\n[INF] %s file has been written out.",outputFileName.c_str());

	end = clock();
	double elapsed_seconds = (double)( end - start ) / (double)CLOCKS_PER_SEC;

	if ( elapsed_seconds < 60) {
		printf ( "\n[INF] Running time was: %.1f seconds\n", elapsed_seconds );
	}
	else {
		int seconds = (int)elapsed_seconds % 60;
		int minutes = ((int)elapsed_seconds-seconds)/60;
		printf ( "\n[INF] Running time was: %i minutes %i seconds.\n", minutes, seconds );
	}
	return 0;
}
